Attribute VB_Name = "modPCodeUtility"
Option Explicit

'#Const bDEBUG = True

'Relevant offsets in the execution buffer
'0FC0D00C+3C.14     = current project object in a global structure
'project.0C         = project code
'projcode.30        = collection of component code buffers (every 30 is obj ptr terminated by FFFFFFFF)
'component.20       = number of procs in component
'component.24       = array of pointers to the proc pcode trailer

'Note that the component.24 array contains an entry for all members, but non-proc members
'such as module-level declarations have a null entry.

'The proc pcode trailer follow the p-code bytes. At a positive offset of 8 from the trailer
'is the integer negative offset of the first p-code byte. This is also the count of p-code bytes
'because the trailer begins immediately after the p-code.

Declare Sub CopyMemory Lib "kernel32" Alias "RtlMoveMemory" (Destination As Any, Source As Any, ByVal Length As Long)
Declare Function GetModuleHandle Lib "kernel32" Alias "GetModuleHandleA" (ByVal lpModuleName As String) As Long
Declare Sub DebugBreak Lib "kernel32" ()

Const COMMENT_STYLE = "'> "
Const VB_RUN_INSTANCE = &HFC0D00C
Const PROJECT_VTABLE = &HFAD0560
Const OPCODE_COUNT = 1351                '00 to FF46

Const OPCODE_DATA_FILE = "opcodes.txt"  'just the name of the file (must be in app.path)

Type OpCodeData
    Byte1 As Byte
    Byte2 As Byte
    VMJump As String * 4
    VBAJump As String * 4
    ArgSize As String * 2
    ArgCount As String * 2
    Arg1Size As String * 2
    Arg2Size As String * 2
    Arg3Size As String * 2
    Instruction As String * 80
    Comment As String * 80
End Type

Dim mOpCodes(1 To OPCODE_COUNT) As OpCodeData
Dim mbOpCodeDataLoaded As Boolean

'Directives:
'"Continuous"       puts disassembled bytes in continuous list before proc
'"Interspersed"     intersperses disassembled bytes with proc source
'"Raw"              puts raw bytes at head of proc

Public Sub ExtractCurrentProcPCode(sDirective As String)

    Dim sCallError As String

    On Error GoTo EH
    
    If Not mbOpCodeDataLoaded Then MsgBox "OpCode data is not loaded.": Exit Sub
    
    'Determine the current member name
    Dim lngSelStartLine As Long, lngStartColumn As Long, lngEndLine As Long, lngEndColumn As Long
    Dim sSelectedProc As String, eProcKind As vbext_ProcKind
    Dim theIDE As vbide.VBE
    Set theIDE = modAddIn.theConnection.theIDE
    theIDE.ActiveCodePane.GetSelection lngSelStartLine, lngStartColumn, lngEndLine, lngEndColumn
    sSelectedProc = theIDE.ActiveCodePane.CodeModule.ProcOfLine(lngSelStartLine, eProcKind)
    
    If sSelectedProc = "" Then MsgBox "You must have the cursor in a function or other body of code to use this command.": Exit Sub

    'Determine the member index of the selected proc, note that not all members have associated code
    'so a separate index is maintained to count the "method/event" members for which there is code
    Dim xMember As Long, ctMembers As Long, eMemberType As vbext_MemberType, xCodeMember As Long
    Dim sMemberList As String
    ctMembers = theIDE.ActiveCodePane.CodeModule.Members.Count
    Do
        xMember = xMember + 1
        If xMember > ctMembers Then MsgBox "Unable to find member named """ & sSelectedProc & """.": Exit Sub
        eMemberType = theIDE.ActiveCodePane.CodeModule.Members(xMember).Type
        If eMemberType = vbext_mt_Event Or eMemberType = vbext_mt_Method Or eMemberType = vbext_mt_Property Then xCodeMember = xCodeMember + 1
        sMemberList = sMemberList & xMember & " " & xCodeMember & " " & eMemberType & " " & theIDE.ActiveCodePane.CodeModule.Members(xMember).Name
        If theIDE.ActiveCodePane.CodeModule.Members(xMember).Name = sSelectedProc Then Exit Do
    Loop
    'MsgBox sMemberList
    
    'Determine the first line of the procedure
    Dim lineProcStart As Long
    lineProcStart = theIDE.ActiveCodePane.CodeModule.ProcBodyLine(sSelectedProc, eProcKind)
    
    'Determine the component index
    Dim theComponent As VBComponent
    Dim xComponent As Long, ctComponents As Long, sComponentName As String
    Set theComponent = theIDE.ActiveCodePane.CodeModule.Parent
    sComponentName = theComponent.Name
    ctComponents = theComponent.Collection.Count
    Do
        xComponent = xComponent + 1
        If xComponent > ctComponents Then MsgBox "Unable to find component named """ & sComponentName & """.": Exit Sub
        If theComponent.Collection(xComponent).Name = sComponentName Then Exit Do
    Loop
    
    'Get P-code bytes
    Dim vPCodeBytes As Variant, sLocaleInfo As String
    If ExtractProcPCode(ctComponents, xComponent, xCodeMember, sLocaleInfo, vPCodeBytes, sCallError) Then
        'we have the p-code
    Else
        MsgBox "Unable to extract proc p-code: " & sCallError & vbNewLine & "Note that you must run the project beforehand to compile it and generate the p-code."
        Exit Sub
    End If
    
    'Group them into instructions
    If sDirective <> "Raw" Then
        Dim bComplete As Boolean, axInstructionOffset() As Long, asInstructionText() As String
        If ParseBytes_Main(vPCodeBytes, axInstructionOffset, asInstructionText, bComplete, sCallError) Then
            'got the instructions
        Else
            MsgBox "Failed to parse p-code byte block: " & sCallError
            Exit Sub
        End If
    End If
    
    'Put instruction text into module
    Dim sOutputLine As String, lineInsertLocation As Long
    lineInsertLocation = lineProcStart
    Select Case sDirective
        Case "Continuous"
            Dim xInstruction As Long
            For xInstruction = axInstructionOffset(0) To 1 Step -1
                theComponent.CodeModule.InsertLines lineInsertLocation, asInstructionText(xInstruction)
            Next
        Case "Interspersed"
            
        Case "Raw"
            Dim xByte As Long, xColumn As Long, aPCodeBytes() As Byte, ctBytes As Long
            aPCodeBytes = vPCodeBytes
            ctBytes = UBound(aPCodeBytes)
            xColumn = 1
            sOutputLine = COMMENT_STYLE & sLocaleInfo & vbNewLine & COMMENT_STYLE
            For xByte = 1 To ctBytes
                xColumn = xColumn + 1
                Cat sOutputLine, Byte2HexString(aPCodeBytes(xByte)) & " "
                If xColumn = 20 Or xByte = ctBytes Then
                    theComponent.CodeModule.InsertLines lineInsertLocation, sOutputLine
                    lineInsertLocation = lineInsertLocation + 1
                    xColumn = 1
                    sOutputLine = COMMENT_STYLE
                End If
            Next
    End Select

    Exit Sub
EH:
    MsgBox "Unexpected error on line " & Erl & " extracting p-code for current proc: " & Err.Description
End Sub

Public Function ExtractProcPCode(ctVBEComponents As Long, ComponentIndex As Long, MemberIndex As Long, sLocaleInfo As String, vByteBox As Variant, sError As String) As Boolean

    On Error GoTo EH

    'Obtain and verify project pointer:
    Dim lpProject As Long
    lpProject = Deref(Deref(VB_RUN_INSTANCE + &H3C) + &H14)
    If Deref(lpProject) = PROJECT_VTABLE Then
        'we are definitely pointing at the project
    Else
        sError = "Unable to locate project execution object."
        GoTo Fail
    End If
    
    'Obtain count of component modules and their pointers
    Dim ctComponents As Integer, lpCCList As Long, xCCList As Long, alpComponentCode() As Long
    ctComponents = DerefInt(Deref(lpProject + &HC) + &H2E)
    If ctComponents <> ctVBEComponents Then sError = "In project at " & Hex(lpProject) & " component list count (" & ctComponents & ") did not match the number of components in the VBE (" & ctVBEComponents & ").": GoTo Fail
    lpCCList = Deref(Deref(lpProject + &HC) + &H30)
    If lpCCList = 0 Then sError = "Component list pointer is null.": GoTo Fail
    ReDim alpComponentCode(1 To ctComponents)
    For xCCList = 1 To ctComponents
        If Deref(lpCCList + &H30 * (xCCList - 1) + 4) <> &HFFFFFFFF Then sError = "Validity check on component pointer (" & Hex(lpCCList) & ") list item " & xCCList & " failed.": GoTo Fail
        alpComponentCode(xCCList) = Deref(lpCCList + (&H30 * (xCCList - 1)))
    Next
    
    'Obtain count of procs in desired component and pointer to desired p-code trailer
    Dim lpComponentCodeObject As Long, ctProcs As Integer, lpDesiredProcTrailer As Long
    If ComponentIndex = 0 Or ComponentIndex > ctComponents Then sError = "Component index " & ComponentIndex & " more than " & ctComponents & " available components.": GoTo Fail
    lpComponentCodeObject = alpComponentCode(ComponentIndex)
    ctProcs = DerefInt(lpComponentCodeObject + &H20)
    If MemberIndex > ctProcs Then sError = "Proc index " & MemberIndex & " more than " & ctProcs & " available procs in component at " & Hex(lpComponentCodeObject) & ".": GoTo Fail
    lpDesiredProcTrailer = Deref(Deref(lpComponentCodeObject + &H24) + 4 * (MemberIndex - 1))
    
    'Validate trailer, and put the code bytes in a byte array
    Dim offCodeBytes As Integer, aCodeBytes() As Byte, xCodeByte As Long
    If lpDesiredProcTrailer = 0 Then sError = "Proc " & MemberIndex & " trailer pointer in component " & Hex(lpComponentCodeObject) & " is null.": GoTo Fail
    If Deref(lpDesiredProcTrailer) <> lpComponentCodeObject Then sError = "Proc " & MemberIndex & " trailer parent reference does not match component address (" & Hex(lpComponentCodeObject) & ").": GoTo Fail
    offCodeBytes = DerefInt(lpDesiredProcTrailer + &H8)
    If offCodeBytes < 1 Then sError = "Count of code bytes less than one.": GoTo Fail
    ReDim aCodeBytes(1 To offCodeBytes)
    For xCodeByte = 1 To offCodeBytes
        aCodeBytes(xCodeByte) = DerefByte(lpDesiredProcTrailer - (offCodeBytes - xCodeByte + 1))
    Next

    'Record where bytes are coming from
    sLocaleInfo = "lpComponent " & ComponentIndex & " of " & ctComponents & ": " & Hex(lpComponentCodeObject) & " lpTrailer: " & Hex(lpDesiredProcTrailer)

    'Stuff the code bytes in a variant for shipment back to caller
    vByteBox = aCodeBytes

    ExtractProcPCode = True
    Exit Function
    
EH:
    sError = "Unexpected error extracting project p-code: " & Err.Description
Fail:
    ExtractProcPCode = False
End Function

'axInstructionOffset    is an array of longs pointing to the first byte of the instruction
'                       axInstructionOffset(0) = the count of instructions
'
'asInstructionText      is an array of strings describing the instruction, the text
'                       has not only generic info about the instruction but also gives
'                       an interpretation of arguments, etc
'
'bComplete              signifies whether the entire body of code bytes was successfully
'                       isolated into instructions. if there was a failure along the way
'                       then the problem bytes will be grouped together and come out as
'                       a glob in the instruction text

Function ParseBytes_Main(vPCodeBytes, axInstructionOffset() As Long, asInstructionText() As String, bComplete As Boolean, sError As String) As Boolean
    
    Dim sCallError As String
    Dim aPCodeBytes() As Byte, xByte As Long, ctByte As Long
    Dim xOpCode As Long, bOpCodeError As Boolean
    Dim aOpCodeNumber() As Long, aArgSz() As Long
    Dim ctInstruction As Long, xInstruction As Long

    On Error GoTo EH
    
    aPCodeBytes = vPCodeBytes
    ctByte = UBound(aPCodeBytes)
    
1    If ParseBytes_CountInstructions(aPCodeBytes, axInstructionOffset, aOpCodeNumber, aArgSz, sCallError) Then
2        ctInstruction = axInstructionOffset(0)
        ReDim asInstructionText(0 To ctInstruction)
3    Else
4        sError = "Failed to count instructions: " & sCallError
5        GoTo Fail
6    End If
7
8    'Make the instruction text
9    Dim xArg As Long, sArgText As String, szArg As Long
10    Dim sInstructionText As String
11    For xInstruction = 1 To ctInstruction
12        sInstructionText = COMMENT_STYLE
13        xOpCode = aOpCodeNumber(xInstruction)
14        Cat sInstructionText, PadLeft(mOpCodes(xOpCode).VBAJump, 6)
15        Cat sInstructionText, PadLeft(OpCodeIndex2String(xOpCode), 7)

        'Build argument string and add it to the instruction text
        xByte = axInstructionOffset(xInstruction)
        If xByte < 1 Then
            sError = "Invalid instruction offset for instruction " & xInstruction
            GoTo Fail
        End If

        sArgText = ""
        If Trim(mOpCodes(xOpCode).ArgSize) = "n" Then 'just append first arg
            Cat sArgText, Byte2HexString(aPCodeBytes(xByte)) & " "
            Cat sArgText, Byte2HexString(aPCodeBytes(xByte + 1)) & " "
            Cat sInstructionText, PadLeft(sArgText, 12)
        Else
            For xArg = 1 To aArgSz(xInstruction)
                If xByte + xArg - 1 > ctByte Then
                    Cat sArgText, "OF " 'overflow
                Else
                    Cat sArgText, Byte2HexString(aPCodeBytes(xByte + xArg - 1)) & " "
                End If
            Next
            If aArgSz(xInstruction) <= 4 Then
                Cat sInstructionText, PadLeft(sArgText, 12)
            Else
                Cat sInstructionText, PadLeft(sArgText, 3 * aArgSz(xInstruction))
            End If
        End If
        Cat sInstructionText, Space(1)                       'additional margin

        Cat sInstructionText, ParseBytes_InstructionInserts(xOpCode, aPCodeBytes, xByte)
        If Len(Trim(mOpCodes(xOpCode).Comment)) > 0 Then Cat sInstructionText, " //" & mOpCodes(xOpCode).Comment & " "

        If Trim(mOpCodes(xOpCode).ArgSize) = "n" Then 'add the list of n-arguments
            Dim sInsertExp As String, posTilde As Long
            posTilde = InStr(mOpCodes(xOpCode).Instruction, " ~ ")
            If posTilde > 0 Then sInsertExp = Right(mOpCodes(xOpCode).Instruction, Len(mOpCodes(xOpCode).Instruction) - posTilde - 2) Else sInsertExp = ""
            For xArg = 3 To aArgSz(xInstruction) Step 2
                Cat sInstructionText, vbNewLine & COMMENT_STYLE & Space(13)
                Cat sInstructionText, Byte2HexString(aPCodeBytes(xByte + xArg - 1)) & " "
                Cat sInstructionText, Byte2HexString(aPCodeBytes(xByte + xArg)) & " "
                Cat sInstructionText, Space(8) 'indent description by 2
                Cat sInstructionText, ParseBytes_InstructionInsert1(sInsertExp, aPCodeBytes, xByte + xArg - 1)
            Next
        End If

        asInstructionText(xInstruction) = sInstructionText

    Next

44    If xByte + aArgSz(ctInstruction) < ctByte Then 'unparsed bytes remain at end
45        bComplete = False
46        Cat asInstructionText(ctInstruction), sCallError & " Remaining bytes: " & vbNewLine & COMMENT_STYLE & Space(6)
47        Dim xColumn As Long, sOutput As String
48        xColumn = 1
49        sOutput = ""
50        For xByte = xByte To ctByte
51            xColumn = xColumn + 1
52            Cat sOutput, Byte2HexString(aPCodeBytes(xByte)) & " "
53            If xByte = ctByte Then Cat asInstructionText(ctInstruction), sOutput
54            If xColumn = 20 Then Cat sOutput, vbNewLine & COMMENT_STYLE & Space(6): xColumn = 1
55        Next
56    Else
57        bComplete = True
58    End If
59
60    ParseBytes_Main = True
    
    Exit Function
EH:
    sError = "Unexpected error on line " & Erl & ": " & Err.Description & "  ctByte: " & ctByte & " xByte: " & xByte & " xOpCode: " & xOpCode & " ctInstruction: " & ctInstruction & " xInstruction: " & xInstruction
Fail:
End Function

'Counts the number of instructions in a set of p-code bytes and returns value as InstructionOffset(0)
'Fills the InstructionOffset array with the starting byte number for the given instruction's arguments
'If the instruction has no argument then the offset is one past the end of the instruction
Function ParseBytes_CountInstructions(aPCodeBytes() As Byte, axInstructionOffset() As Long, aOpCodeNumber() As Long, aArgSz() As Long, sError As String) As Boolean
    Dim sCallError As String, xByte As Long, ctBytes As Long, xInstruction As Long
    Dim lngOCNumber As Long
    On Error GoTo EH
    ctBytes = UBound(aPCodeBytes)
    ReDim axInstructionOffset(0 To ctBytes)
    ReDim aOpCodeNumber(0 To ctBytes)
    ReDim aArgSz(0 To ctBytes)
    xByte = 1: xInstruction = 0
    Do
        If xByte > ctBytes Then Exit Do
        If ParseBytes_GetOCNumber(aPCodeBytes, xByte, lngOCNumber, sCallError) Then
            xInstruction = xInstruction + 1
            axInstructionOffset(xInstruction) = xByte
            aOpCodeNumber(xInstruction) = lngOCNumber
            If IsNumeric(mOpCodes(lngOCNumber).ArgSize) Then
                aArgSz(xInstruction) = CLng(mOpCodes(lngOCNumber).ArgSize)
            Else
                If Trim(mOpCodes(lngOCNumber).ArgSize) = "n" Then 'there are n args
                    Select Case Trim(mOpCodes(lngOCNumber).ArgCount)
                        Case "n"
                            aArgSz(xInstruction) = 2 + Bytes2Int(aPCodeBytes, xByte) * 2
                        Case "/2"  'arg count = imm1#2/2, arg size = imm1#2
                            aArgSz(xInstruction) = 2 + Bytes2Int(aPCodeBytes, xByte)
                        Case Else
                            aArgSz(xInstruction) = 0    'not a recognized count
                            Exit Do                     'abort count
                    End Select
                Else
                    aArgSz(xInstruction) = 0    'no arg size defined
                    Exit Do                     'abort count
                End If
            End If
            xByte = xByte + aArgSz(xInstruction)
        Else
            sError = "Failed to get opcode number for byte " & xByte & " instruction " & xInstruction & ": " & sCallError
            Exit Do     'stop counting
        End If
    Loop
    axInstructionOffset(0) = xInstruction                   'the instruction count
    ReDim Preserve axInstructionOffset(0 To xInstruction)
    ReDim Preserve aOpCodeNumber(0 To xInstruction)
    ReDim Preserve aArgSz(0 To xInstruction)
    ParseBytes_CountInstructions = True
    Exit Function
EH:
    sError = "Unexpected error counting instructions: " & Err.Description
Fail:
End Function

'Determine the opcode number (from 1 to 1351)
'Automatically advances byte index by reference
Function ParseBytes_GetOCNumber(aPCodeBytes() As Byte, xByte As Long, lngOCNumber As Long, sError As String) As Boolean
    If xByte > UBound(aPCodeBytes) Then
        sError = "Byte number " & xByte & " past end.": GoTo Fail
    Else
        If xByte < 1 Then
            sError = "Byte number is zero.": GoTo Fail
        Else
            'byte number is valid
        End If
    End If
    If aPCodeBytes(xByte) < &HFB Then
        lngOCNumber = CLng(aPCodeBytes(xByte)) + 1      'the sequential index of the opcode (from 1 to 1351)
        xByte = xByte + 1                               'advance past opcode
    Else 'escape byte
        If xByte + 1 > UBound(aPCodeBytes) Then sError = "P-Code group terminated by an escape byte.": GoTo Fail
        If aPCodeBytes(xByte) = &HFF And aPCodeBytes(xByte + 1) > &H46 Then sError = "Double byte code invalid. Greater than 0xFF46.": GoTo Fail
        lngOCNumber = (aPCodeBytes(xByte) - &HFA) * &H100 + aPCodeBytes(xByte + 1)
        xByte = xByte + 2                           'advance past opcode
    End If
    ParseBytes_GetOCNumber = True
    Exit Function
EH:
    sError = "Unexpected error getting opcode number: " & Err.Description
Fail:
End Function

Function ParseBytes_InstructionInserts(xOpCode As Long, aPCodeBytes() As Byte, xByte As Long) As String
    Dim sInstructionText As String
    Dim posTilde As Long
    On Error GoTo EH
    sInstructionText = mOpCodes(xOpCode).Instruction
    If Trim(mOpCodes(xOpCode).ArgSize) = "n" Then 'remove ~ portion of instruction and add a colon
        posTilde = InStr(sInstructionText, " ~ ")
        If posTilde > 0 Then sInstructionText = Left$(sInstructionText, posTilde - 1) & ":"
    End If
    If IsNumeric(mOpCodes(xOpCode).ArgCount) Then
        If CInt(mOpCodes(xOpCode).ArgCount) = 1 Then
            If InStr(sInstructionText, "imm") > 0 Then
                Select Case CInt(mOpCodes(xOpCode).Arg1Size)
                    Case 2
                        sInstructionText = Replace(sInstructionText, "imm", Bytes2HexInt(aPCodeBytes, xByte))
                    Case 4
                        sInstructionText = Replace(sInstructionText, "imm", Bytes2HexInt(aPCodeBytes, xByte))
                End Select
            Else
                If InStr(sInstructionText, "arg") > 0 Then
                    sInstructionText = Replace(sInstructionText, "arg", ParseBytes_TranslateArg(aPCodeBytes, xByte))
                End If
            End If
        Else         'multiple args
        End If
    End If
    ParseBytes_InstructionInserts = sInstructionText
    Exit Function
EH:
    ParseBytes_InstructionInserts = "[Error determining instruction text: " & Err.Description & "]"
End Function

'Inserts address expressions for a single two-byte arg or ptr
Function ParseBytes_InstructionInsert1(sInstructionText As String, aPCodeBytes() As Byte, xByte As Long) As String
    On Error GoTo EH
    If InStr(sInstructionText, "arg") > 0 Then
        ParseBytes_InstructionInsert1 = Replace(sInstructionText, "arg", ParseBytes_TranslateArg(aPCodeBytes, xByte))
    Else
        ParseBytes_InstructionInsert1 = sInstructionText
    End If
    Exit Function
EH:
    ParseBytes_InstructionInsert1 = "[Error doing insert1: " & Err.Description & "]"
End Function

Function ParseBytes_TranslateArg(aPCodeBytes() As Byte, xByte As Long) As String
    Dim intArg As Integer
    intArg = CInt("&H" & Bytes2HexInt(aPCodeBytes(), xByte))
    If intArg > 0 Then 'stack
        ParseBytes_TranslateArg = "stack." & Hex(intArg)
    Else 'local variable
        intArg = intArg * -1
        ParseBytes_TranslateArg = "local." & Hex(intArg)
    End If
End Function

Function Bytes2Int(aPCodeBytes() As Byte, xByte As Long) As Long
    Bytes2Int = aPCodeBytes(xByte + 1) * 256 + aPCodeBytes(xByte)
End Function

Function Bytes2HexInt(aPCodeBytes() As Byte, xByte As Long) As String
    If xByte + 1 > UBound(aPCodeBytes) Then Bytes2HexInt = "ERR": Exit Function
    Bytes2HexInt = Byte2HexString(aPCodeBytes(xByte + 1)) & Byte2HexString(aPCodeBytes(xByte))
End Function

Function Bytes2HexLong(aPCodeBytes() As Byte, xByte As Long) As String
    If xByte + 3 > UBound(aPCodeBytes) Then Bytes2HexLong = "ERR": Exit Function
    Bytes2HexLong = Bytes2HexInt(aPCodeBytes, xByte + 2) & Bytes2HexInt(aPCodeBytes, xByte)
End Function

Function Deref(lngPointer As Long) As Long  'Equivalent of *lngPointer (returns the value pointed to)
    Dim lngValueAtPointer As Long
    CopyMemory lngValueAtPointer, ByVal lngPointer, 4
    Deref = lngValueAtPointer
End Function

Function DerefInt(lngPointer As Long) As Integer  'Equivalent of *intPointer (returns the value pointed to)
    Dim intValueAtPointer As Integer
    CopyMemory intValueAtPointer, ByVal lngPointer, 4
    DerefInt = intValueAtPointer
End Function

Function DerefByte(lngPointer As Long) As Byte  'Equivalent of *intPointer (returns the value pointed to)
    Dim byteValueAtPointer As Byte
    CopyMemory byteValueAtPointer, ByVal lngPointer, 1
    DerefByte = byteValueAtPointer
End Function

'This is the format for the Opcode.Txt file
' - lines starting with a ; are treated as a comment
' - blank lines are ignored
' - the byte codes do not have to be in order

'Note that there is some spacing between the fields

Public Function LoadOpCodeData(sError As String) As Boolean
    Dim sPath As String, hFile As Integer, bFileOpen As Boolean
    Dim xLine As Long, xOpCode As Long
    Dim Byte1 As Byte, Byte2 As Byte, sByte2 As String * 2
    Dim sCurrentLine As String
    On Error GoTo EH
    sPath = App.Path & "\" & OPCODE_DATA_FILE
    hFile = FreeFile
    If Dir(sPath) = "" Then sError = "Required file not found in app directory: " & sPath: GoTo Fail
    Open sPath For Input As hFile
    bFileOpen = True
    Do While Not EOF(hFile)
        xLine = xLine + 1
        Line Input #hFile, sCurrentLine
        If Trim(sCurrentLine) = "" Then GoTo Continue
        If Left$(sCurrentLine, 1) = ";" Then GoTo Continue
        If Not IsHexByte(Left(sCurrentLine, 2), Byte1) Then sError = "Chars 1 and 2 on line " & xLine & " (" & Left(sCurrentLine, 2) & " do not constitute a valid hex byte.": GoTo Fail
        sByte2 = Mid$(sCurrentLine, 4, 2)
        If Trim(sByte2) = "" Then             'single byte opcode
            xOpCode = CLng(Byte1) + 1   'the sequential index of the opcode (from 1 to 1351)
        Else
            If Not IsHexByte(sByte2, Byte2) Then sError = "Byte2 on line " & xLine & " is not a valid hex byte.": GoTo Fail
            If Byte1 = &HFF And Byte2 > &H46 Then sError = "Byte code on line " & xLine & " invalid. Greater than 0xFF46.": GoTo Fail
            If Byte1 < &HFB Then sError = "Byte code on line " & xLine & " invalid. Escape byte is less than 0xFB.": GoTo Fail
            xOpCode = (Byte1 - &HFA) * &H100 + Byte2
        End If
        With mOpCodes(xOpCode)                          'Start  Width   Item
            .Byte1 = Byte1                              '1      2       Byte1
            .Byte2 = Byte2                              '4      2       Byte2
            .VMJump = Mid$(sCurrentLine, 7, 4)          '7      8       VM Jump
            .VBAJump = Mid$(sCurrentLine, 12, 4)        '12     8       VBA Jump
            .ArgSize = Mid$(sCurrentLine, 17, 2)        '17     2       Arg Size
            .ArgCount = Mid$(sCurrentLine, 19, 2)       '19     2       Arg Count
            .Arg1Size = Mid$(sCurrentLine, 21, 2)       '21     2       Arg1 Size
            .Arg2Size = Mid$(sCurrentLine, 23, 2)       '23     2       Arg2 Size
            .Arg3Size = Mid$(sCurrentLine, 25, 2)       '25     2       Arg3 Size
            .Instruction = Mid$(sCurrentLine, 28, 80)   '28     80      Instruction
            .Comment = Mid$(sCurrentLine, 110, 80)      '110    80      Comment
        End With
Continue:
    Loop
    If bFileOpen Then Close hFile
    LoadOpCodeData = True
    mbOpCodeDataLoaded = True
    Exit Function
EH:
    sError = "Unexpected error while loading data line " & xLine & " : " & Err.Description
Fail:
    If bFileOpen Then Close hFile
    LoadOpCodeData = False
End Function

Sub Cat(sString, sTextToAppend)
    sString = sString & sTextToAppend
End Sub

Function IsHexByte(sHex As String, byteValue As Byte) As Boolean
    On Error GoTo EH
    byteValue = CByte("&H" & sHex)
    IsHexByte = True
    Exit Function
EH:
    IsHexByte = False
End Function

Function OpCodeIndex2String(ByVal xOpCode As Long) As String
    Select Case xOpCode
        Case 1 To 256
            OpCodeIndex2String = Long2HexString(xOpCode - 1)
        Case 257 To 512
            OpCodeIndex2String = "FB " & Long2HexString(xOpCode - 256)
        Case 513 To 768
            OpCodeIndex2String = "FC " & Long2HexString(xOpCode - 512)
        Case 769 To 1024
            OpCodeIndex2String = "FD " & Long2HexString(xOpCode - 768)
        Case 1025 To 1280
            OpCodeIndex2String = "FE " & Long2HexString(xOpCode - 1024)
        Case 1281 To 1351
            OpCodeIndex2String = "FF " & Long2HexString(xOpCode - 1280)
        Case Else
            OpCodeIndex2String = "???"
    End Select
End Function

Function Byte2HexString(byt As Byte) As String
    On Error GoTo EH
    If byt < 16 Then
        Byte2HexString = "0" & Hex(byt)
    Else
        Byte2HexString = Hex(byt)
    End If
    Exit Function
EH:
    Byte2HexString = "??"
End Function

Function Long2HexString(lng As Long) As String
    On Error GoTo EH
    If lng < 16 Then
        Long2HexString = "0" & Hex(lng)
    Else
        Long2HexString = Hex(lng)
    End If
    Exit Function
EH:
    Long2HexString = "??"
End Function

'Makes a string into a fixed length left-aligned
Function PadLeft(sString, lngSize) As String
    Select Case Len(sString)
        Case Is < lngSize
            PadLeft = sString & Space(lngSize - Len(sString))
        Case Is = lngSize
            PadLeft = sString
        Case Is > lngSize
            PadLeft = Left(sString, lngSize)
    End Select
End Function
